/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* cis_base.c

   Connected induced subgraphs
   Ambros Marzetta

   $Id: cis_base.c 5167 2004-10-20 16:20:33Z bremner $
 */


#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include "z_base.h"
#include "z_rs.h"
#include "z_enumps.h"
#include "z_par.h"		/* z_myproc */
#include "z_time.h"
#include "z_adjlist.h"
#include "z_ps2d.h"
#include "cis_base.h"

static boolean cis_base_inited;

int cis_NodeTag, cis_GlobalTag;

cis_NodeP cis_NewNode(cis_pointT numPt, cis_pointT actPt)
{
  cis_NodeP n;

  assert(actPt <= numPt);
  n = (cis_NodeP) z_NewNode(cis_NodeTag,
			    sizeof(cis_Node) + numPt * sizeof(boolean));
  n->numAct = actPt;
  return n;
} /* cis_NewNode */

void cis_FreeNode(cis_NodeP node)
{
  z_FreeNode((z_NodeP) node);
} /* cis_FreeNode */

boolean cis_Equal(cis_GlobalP global, cis_NodeP x, cis_NodeP y)
{
  int i;

  assert(global && global->tag == cis_GlobalTag);
  assert(x);
  assert(y);
  if (x->numAct != y->numAct) return 0;
  for (i=0; i<global->numPt; i++)
    if (x->pt[i] != y->pt[i]) return 0;
  return 1;
} /* cis_Equal */

void cis_Print(cis_GlobalP global, cis_NodeP node)
{
  int i;

  assert(global && global->tag == cis_GlobalTag);
  printf("[%2d][", z_myproc);
  if (!node)
    printf("NULL");
  else
    for (i=0; i<global->numPt; i++)
      if (node->pt[i]) printf("%3d", i);
  printf("]\n");
} /* cis_Print */

#define BILLION 1000000000
#define INSIDE_FRACTION (1/12.0) /* distance from outermost point to border */

static void bound(ps2d_coordT x, ps2d_coordT *min, ps2d_coordT *max)
{
  /* Enlarge the interval [min..max] such that x is in the interval */
  if (x < *min) *min = x;
  if (x > *max) *max = x;
} /* bound */

void cis_PSHeader(cis_GlobalP global, FILE *psfile, float xsize, float ysize)
{
  ps2d_coordT xmin, xmax, ymin, ymax;
  int i;
  float ixsize, iysize;
  
  assert(global && global->tag == cis_GlobalTag);
  xmin = BILLION;
  xmax = -BILLION;
  ymin = BILLION;
  ymax = -BILLION;
  for (i=0; i<global->numPt; i++) {
    bound(cis_coord_x(global, i), &xmin, &xmax);
    bound(cis_coord_y(global, i), &ymin, &ymax);
  }
  printf("%%bounds: [%f %f] [%f %f]\n", xmin, xmax, ymin, ymax);
  if (xmax == xmin) xmax ++;	/* to avoid division by zero */
  if (ymax == ymin) ymax ++;
  ixsize = xmax - xmin;		/* final coordinate system */
  iysize = ymax - ymin;

  fprintf(psfile, "/radius %f def /radius3 radius 3 mul def\n",
	  1 / xsize * ixsize);
  fprintf(psfile, "/points {\n");
  fprintf(psfile, "  %f setlinewidth\n", 0.3 * ixsize / xsize /
	  (1 + 2 * INSIDE_FRACTION));
  fprintf(psfile, "  %f %f scale\n",
	  1 / ixsize * (xsize) / (1 + 2 * INSIDE_FRACTION),
	  1 / iysize * (ysize) / (1 + 2 * INSIDE_FRACTION));
  fprintf(psfile, "  %f %f translate\n",
	  ixsize * INSIDE_FRACTION - xmin , iysize * INSIDE_FRACTION - ymin);
  for (i=0; i<global->numPt; i++) {
    fprintf(psfile, "  newpath %f %f radius 0 360 arc fill\n",
	    cis_coord_x(global, i),
	    cis_coord_y(global, i));
  }
  fprintf(psfile, "} def\n");
} /* cis_PSHeader */

void cis_PSPicture(cis_GlobalP global, cis_NodeP node, FILE *psfile,
		   float xsize, float ysize)
{
  int pt;
  cis_edgeT *adj;
  
  assert(global && global->tag == cis_GlobalTag);
  fprintf(psfile, "points\n");	/* defined in cis_PSHeader */
  for (pt=0; pt<global->numPt; pt++) {
    if (node->pt[pt]) {
      fprintf(psfile, "newpath %f %f radius3 0 360 arc fill\n",
	      cis_coord_x(global, pt),
	      cis_coord_y(global, pt));
      adj = cis_adjPt(global, pt);
      while (*adj != ADJ_LAST) {
	if (node->pt[*adj]) {
	  if (*adj > pt)
	    fprintf(psfile, "newpath %f %f moveto %f %f lineto stroke\n",
		    cis_coord_x(global, pt), cis_coord_y(global, pt),
		    cis_coord_x(global, *adj), cis_coord_y(global, *adj));
	}
	adj++;
      } /* while */
    } /* if pt */
  } /* for pt */
} /* cis_PSPicture */

/* ---------------- */

#define MAXHEADERLINE 128

cis_GlobalP cis_Read(FILE *infile)
     /* Read Graph from infile and allocate a full node */
     /* Should test whether graph is connected. */
     /* Should test for parallel edges */
{
  char s[MAXHEADERLINE], c;
  int points, edges, actpt, line, i, index, matches;
  cis_GlobalP global;

  points = 0;
  edges = 0;
  line = 0;			/* line numbers start with 1 */
  do {				/* read header line */
    (void) fgets(s, MAXHEADERLINE, infile);
    line ++;
    matches = sscanf(s, " Points:%d", &points);
    if (matches == 0) {
      if (sscanf(s, " Edges:%d", &edges) <= 0) 
	if (sscanf(s, " %c", &c) == 1 && c != '#') {
	  fprintf(stderr, "cis_Read: Error in line %1d\n", line);
	  return NULL;
	}
    } else if (feof(infile)) {
      fprintf(stderr, "cis_Read: End of file before line %d\n", line);
      return NULL;
    }
  } while (points==0 || edges==0);
  
  actpt = 0; /* points with positive degree */
  global = (cis_GlobalP)z_NewNode(cis_GlobalTag, sizeof(cis_Global));
  if (!global) fprintf(stderr, "cis_Read: Could not allocate global.\n");
  global->numPt = points;
  global->cis_adjlist = z_adjlist_New(points, edges);
  global->cis_coord = z_ps2d_New(points);
  z_ps2d_ReadStart(global->cis_coord);
  z_adjlist_ReadStart(global->cis_adjlist, &i, &index);
  for (;;) {
    if (fscanf(infile, "%1s", &c) != 1)
      break;
    switch(c) {
    case 'P':
      z_adjlist_Read(infile, global->cis_adjlist, &i, &index, MAXDEGREE);
      break;
    case 'C':
      z_ps2d_Read(infile, global->cis_coord);
      break;
    default:
      fprintf(stderr, "cis_Read: Illegal character '%c'\n", c);
    } /* switch(ch) */
  } /* for */
  z_adjlist_ReadEnd(global->cis_adjlist, &i, &index);
  return global;
} /* cis_Read */

cis_NodeP cis_EmptySubgraph(cis_GlobalP global)
{
  cis_NodeP root;
  int i;			/* must be signed */

  root = cis_NewNode(global->numPt, 0);
  for (i = global->numPt - 1; i >= 0; --i)
    root->pt[i] = 0;
  return root;
} /* cis_EmptySubgraph */

static boolean cut[MAXDEGREE];	/* output of FindCutPoints */

static void FindCutPoints(cis_GlobalP global, cis_NodeP node)
{
  /* result (set of cut points of node) is in static array cut */
  cis_pointT back[MAXDEGREE], minback[MAXDEGREE], prenum[MAXDEGREE];
  cis_edgeT *adj[MAXDEGREE];
  cis_pointT root, x, other, count;
  int i;			/* must be signed */
  int rootsons;			/* number of sons of root */

  for (i=global->numPt; i>=0; --i) {
    adj[i] = cis_adjPt(global, i);
    back[i] = MAXDEGREE;	/* unvisited points */
    minback[i] = MAXDEGREE;
    cut[i] = 0;
  }
  if (node->numAct <= 2) return; /* no cut points in small graph */
  count = 0;			/* for preorder enumeration */
  root = 0;
  while (!node->pt[root]) root++;
  x = root;
  back[x] = x;			/* mark root as visited */
  prenum[x] = count++;
  rootsons = 0;
  for (;;) {			/* process edge (x, *adj[x]) */
    other = *adj[x];
    adj[x] ++;
    if (other != ADJ_LAST) {
      if (node->pt[other]) {
	if (back[other] == MAXDEGREE) { /* forward edge */
	  if (x == root) rootsons ++; /* this is another son of the root */
	  back[other] = x;
	  prenum[other] = count++; /* preorder number */
	  x = other;
	} else {		/* backward edge or inverse forward edge*/
	  if (prenum[other] < minback[x]) minback[x] = prenum[other];
	}
      }
    } else {			/* no more edges from x; backtrack */
      if (x != root) {		/* not root of depth first tree */
	if (back[x] != root && minback[x] >= prenum[back[x]]) {
	  cut[back[x]] = 1;
	} else if (minback[x] < minback[back[x]]) /* inherit minback */
	  minback[back[x]] = minback[x];
	x = back[x];
      } else
	break;			/* end of depth first search */
    }
  } /* for */

  if (rootsons >= 2) cut[root] = 1;
  /* printf("Cut points: ");
  for (i=0; i<=global->numPt; i++)
    if (cut[i]) printf("%2d", i);
  printf("\n");*/
} /* FindCutPoints */

cis_NodeP cis_adj_oracle(cis_GlobalP global, cis_NodeP node,
				uint32 k)
{
  /* Allocate neighbour k of node, if it exists. This oracle is not
     symmetric, but maybe more efficient than a symmetric one. Include
     point k if the graph remains connected. Exclude k if it is the
     smallest point to exclude such that the graph remains connected.
     Return NULL if neighbour k does not exist.
     k must be in [0..global->numPt-1] */
  
  cis_NodeP oracle = NULL;
  cis_edgeT *adj;
  int i;
  
  assert(k < global->numPt);
  if (node->pt[k]) {		/* exclude point k */
    return NULL;		/* these are never children */
    FindCutPoints(global, node);
    if (!cut[k]) {
      i = 0;
      while (!node->pt[i] || cut[i]) i++;
      if (i == k) {
	oracle = (cis_NodeP) z_NewCopyNode((z_NodeP) node);
	oracle->pt[k] = 0;
	oracle->numAct --;
      }
    } else
      oracle = NULL;		/* new graph is not connected */
  } else {			/* include point k */
    if (node->numAct == 0) {	/* include in empty graph */
      oracle = (cis_NodeP) z_NewCopyNode((z_NodeP) node);
      oracle->pt[k] = 1;
      oracle->numAct = 1;
    } else {
      adj = cis_adjPt(global, k);
      while (*adj != ADJ_LAST && !node->pt[*adj]) adj++;
      if (*adj != ADJ_LAST) {	/* new graph is connected */
	oracle = (cis_NodeP) z_NewCopyNode((z_NodeP) node);
	oracle->pt[k] = 1;
	oracle->numAct ++;
      } else 			/* new graph is not connected */
	oracle = NULL;
    }
  }
  return oracle;
} /* cis_adj_oracle */

cis_NodeP cis_f(cis_GlobalP global, cis_NodeP node)
{
  /* Allocate f(node).
     f(node) = remove smallest point such that the graph remains
               connected.
   */
  cis_pointT x;
  cis_NodeP f_node;

  assert(!is_root_node(node));		/* graph must not be empty */
  FindCutPoints(global, node);
  x = 0;
  while (!node->pt[x] || cut[x]) x++;
  /* x = first non-cut point */
  f_node = (cis_NodeP) z_NewCopyNode((z_NodeP) node);
  f_node->pt[x] = 0;
  f_node->numAct --;
  return f_node;
} /* cis_f */

boolean cis_is_root_node(cis_GlobalP global, cis_NodeP this)
{
  return is_root_node(this);
}

uint32 cis_Delta(cis_GlobalP global, cis_NodeP this)
{
  return global->numPt;
}

void cis_base_Init(void)
{
  /* This function is called once on every processing element. */

  if (! cis_base_inited) {
    cis_base_inited = 1;
    z_ps2d_Init();
    z_adjlist_Init();
    
    cis_NodeTag = z_InstallClass(cis_Node, 0, 0, 0);
    cis_GlobalTag = z_InstallClass(cis_Global, 2, 0, 0);
  }
} /* cis_base_Init */
