/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* z_md_para.c (Paragon machine dependent functions)

   Ambros Marzetta 94-11-22

 * $Source$
 * $Revision: 5167 $
 * $Date: 2004-10-20 13:20:33 -0300 (Wed, 20 Oct 2004) $
 * $Log$
 * Revision 1.3  2004/10/20 16:20:32  bremner
 * added copyright notice and GPL boilerplate
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.10  1996/07/30  15:30:44  marzetta
 * Removed nx_pri() call for new operating system version.
 *
 * Revision 1.9  1996/05/17  13:19:14  marzetta
 * Proxy process catches signals.
 *
 * Revision 1.8  1996/04/09  13:21:05  marzetta
 * Corrected bug when sending to oneself
 *
 * Revision 1.7  1995/09/14  12:26:10  marzetta
 * optimized.
 *
 * Revision 1.6  1995/08/22  12:39:45  marzetta
 * New message passing, MPI
 *
 * Revision 1.4  1995/07/28  11:16:05  marzetta
 * struct buffer_header
 *
 * Revision 1.3  1995/07/28  08:36:28  marzetta
 * z_md_init with path parameter
 *
 * Revision 1.2  1995/07/27  11:36:00  marzetta
 * Separation of message passing interface
 *
 */

#include <nx.h>
#include <sys/types.h>		/* for kill */
#include <signal.h>		/* for kill */
#include <stdio.h>
#include "z_md.h"
#include "z_par_intern.h"

const boolean z_md_doesTD = 0;

/* Message Types ------------------------------------------------------

  -1: MT_ANY, any of the other (not used)
   0: MT_GEN, general
   1: MT_START_JOB, start a parallel job
   2: MT_COLLECT_RESULT
   */

#define MT_ANY -1
#define MT_GEN 0
#define MT_START_JOB 1
#define MT_COLLECT_RESULT 2

#define PTYPE0 0

#define Mega 1000000

long z_memsize = 18 * Mega;	/* Paragon:	 32 MByte real memory */
long z_memavail = 18 * Mega;	/* Paragon:	 32 MByte real memory */

/* Send / Receive ------------------------------------------------ */

#define BUFSIZE 400000
const long z_bufunits = BUFSIZE;
  
static int send_mid = -1;	/* send_mid is valid if (! cansend) */

boolean z_md_TestSend(void)
{
  /* (PE only) Is it possible to overwrite outbuffer? */

  return msgdone(send_mid);
} /* z_md_TestSend */

void z_md_Isend(void *buffer, long size, int dest, char *comment)
{
  /* Send outbuffer to dest*/
  /*printf("[%2d]csend %d\n", z_myproc, outpos * sizeof(unitT), dest);*/
  send_mid = isend(MT_GEN, buffer, size, dest, PTYPE0);
} /* z_md_Isend */

void z_md_Ibroadcast(void *buffer, long size, char *comment)
{
  /* Broadcast outbuffer */
  send_mid = isend(MT_GEN, buffer, size, -1, PTYPE0);
} /* z_md_Ibroadcast */

static long id;

/* Starting jobs ---------------------------------------------------- */

void z_md_StartParallel(void)
{
  /* Broadcast outbuffer and initiate the parallel computation. */
  if (z_numprocs > 1)
    csend(MT_START_JOB, outbuffer,
	  outpos * sizeof(unitT) + sizeof(struct buffer_header), -1, PTYPE0);
} /* z_md_StartParallel */

void z_md_RecBuffer(void)
{
  /* Called from main */
  /* It is possible that a message from another PE arrives here before
     the broadcast from StartParallel. So we need the message type
     MT_START_JOB. */
  crecv(MT_START_JOB, inbuffer,
	BUFSIZE * sizeof(unitT) + sizeof(struct buffer_header));
} /* z_md_RecBuffer */

/* Ending jobs --------------------------------------------------------- */

void z_md_Wait(z_ResultProc resultProc)
{
  int i, numcollected, sender;

  assert(z_myproc == 0);
  if (resultProc) {
    numcollected = 1;

    while (numcollected < z_numprocs) {
      crecv(MT_COLLECT_RESULT, inbuffer,
	    BUFSIZE * sizeof(unitT) + sizeof(struct buffer_header));
      sender = infonode();
      inpos = 0;
      /*printf("Got result from PE %d\n", sender);*/
      resultProc(sender);
      numcollected ++;
    }
  }
} /* z_md_Wait */

void z_md_SendToFE(boolean collect)
/* (PE only) Called by z_pe_switch if !z_frontend.
   Send the result of the computation to the front end which receives
   it in z_md_Wait.
   */
{
  if (collect)			/* send result back to front end */
    csend(MT_COLLECT_RESULT, outbuffer,
	  outpos * sizeof(unitT) + sizeof(struct buffer_header),
	  0, 0);
} /* z_md_SendToFE */


/* Message passing ---------------------------------------------------- */

void z_md_Irecv(void)
{
  long oldid;

  oldid = id;
  id = irecv(MT_GEN, inbuffer,
	     BUFSIZE * sizeof(unitT) + sizeof(struct buffer_header));
#ifdef TRACERECEIVE
  printf("[%2d]#StartReceive %ld%5ld\n", z_myproc, id, oldid);
#endif
} /* z_md_Irecv */

boolean z_md_TestRecv(void)
{
  return msgdone(id);
} /* z_md_TestRecv */

boolean z_md_WaitRecv(void)
/* The processing element is idle.
   There is a receive pending.
   Wait for the receive operation started by z_StartReceive to complete.
   Return true if/when all processing elements are idle.
   Return false if/when a message is received.
   Post: There is no receive pending.
   */
{
  msgwait(id);
  return 0;
} /* z_md_WaitRecv */


/* Assert ---------------------------------------------------------- */

void z_md_abort(void)
{
  abort();
} /* z_md_abort */


/* Signal handler ----------------------------------------------------- */

static void proxy_handler(int signal)
{
  printf("\n[pr]Got signal %d (time=%f) (ignored)\n",
	 signal, z_Wtime());
  fflush(stdout);
  /* Is it allowed to call fflush from a signal handler?
     Is fflush reentrant? */
} /* z_term_handler */


/* Initialization --------------------------------------------------- */

void z_md_Init(int *argc, char ***argv)
     /* Initialize this module. Called by main(). */
{
  /* called by the proxy process */
  int   n, i;
  pid_t pids[MAX_PE];
  static struct sigaction action, o_action;

  /* create application */
  n = nx_initve("", 0, NULL, argc, *argv);
  if(n == -1) {
    /* nx_initve() failed */
    perror("zram.nx_initve");
    exit(1);
  }
  
#if 0
  /* set application priority to 7 */
  n = nx_pri(0, 7);		/* 0 specifies calling application */
  if(n == -1) {
    /* nx_pri() failed */
    perror("nx_pri");
    exit(1);
  }
#endif
  
  /* fork child processes onto all nodes of application */
  n = nx_nfork(NULL, -1, 0, pids);
  if (n == -1) {
    /* nx_nfork() failed */
    perror("nx_nfork");
    exit(1);
  } else if(n == 0) {
    /* child process */
  } else {
    /* parent (controlling process): wait for all children */
    /* proxy process */
    /* ignore signals caught by node processes */
    action.sa_handler = proxy_handler;
    if (sigaction(SIGALRM, &action, &o_action))
      perror("sigaction");
    if (sigaction(SIGTERM, &action, &o_action))
      perror("sigaction");
    nx_waitall();
    exit(0);
  }
  
  /* z_numprocs and z_myproc have not yet their value here */
  inbuffer = (struct buffer *)
    malloc(sizeof(struct buffer_header) + BUFSIZE * sizeof(unitT));
  outbuffer = (struct buffer *)
    malloc(sizeof(struct buffer_header) + BUFSIZE * sizeof(unitT));
  z_par_Init(mynode(), numnodes(), mynode() == 0);
} /* z_md_Init */

void z_md_Fin(void)
/* Called my main() */
{
} /* z_md_Fin */
