/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* z_rs.c (ZRAM: Reverse search)
   Ambros Marzetta

 * $Source$
 * $Revision: 5167 $
 * $Date: 2004-10-20 13:20:33 -0300 (Wed, 20 Oct 2004) $
 * $Log$
 * Revision 1.3  2004/10/20 16:20:32  bremner
 * added copyright notice and GPL boilerplate
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.2  1997/10/03  15:36:28  marzetta
 * Added a type cast to eliminate a compiler warning
 *
 * Revision 1.1  1996/06/03  19:02:19  marzetta
 * Initial revision
 *
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "z_rs.h"
#include "z_ps.h"
#include "z_lb.h"

/* Types --------------------------------------------------------------- */

static int rs_GlobalTag;

typedef struct rs_Global {
  Z_NODE_HEADER;		/* len; tag = rs_GlobalTag */
  z_NodeP rs_app_global;	/* application global node */
  int rs_psn;			/* application procedure set (z_rs_ProcSetP) */
} rs_Global, *rs_GlobalP;

static rs_GlobalP NewRsGlobal(z_ProcSetNum procsetnum, z_lb_GlobalP ap_global)
{
  rs_GlobalP rs_global;

  assert(ap_global);
  rs_global = (rs_GlobalP) z_NewNode(rs_GlobalTag, sizeof(rs_Global));
  rs_global->rs_psn = procsetnum;
  rs_global->rs_app_global = ap_global;
  return rs_global;
} /* NewRsGlobal */


static int z_rs_WorkUnitTag;

typedef struct z_rs_WorkUnit {
  Z_NODE_HEADER;
  z_NodeP start;		/* 1 child */
  uint32 nr_end;
  int32 depth_start, depth_common;
} *z_rs_WorkUnitP;

static z_rs_WorkUnitP NewWorkUnit(z_NodeP start, uint32 nr_end,
				  int32 depth_start, int32 depth_common)
{
  z_rs_WorkUnitP wu;
  wu = (z_rs_WorkUnitP) z_NewNode(z_rs_WorkUnitTag,
				  sizeof(struct z_rs_WorkUnit));
  wu->start = start;
  wu->nr_end = nr_end;
  wu->depth_start = depth_start;
  wu->depth_common = depth_common;
  return wu;
} /* NewWorkUnit */


/* --------------------------------------------------------------------- */

void z_rs_Fini(void);
void z_rs_Next(rs_GlobalP rs_global, z_rs_WorkUnitP *workUnitH);
void z_rs_Split(rs_GlobalP rs_global,
		z_rs_WorkUnitP workUnit,
		s_seqP workList);

#define app_global (rs_global->rs_app_global)

static z_lb_ProcSet lb_procset;
static int lb_procsetnum;
static boolean rs_inited;
static z_NodeP cacheTop, cacheBottom;
static int cacheSize;
static long cacheHits, cacheMisses;
uint32 z_rs_cacheSize = 40;


static z_NodeP NextChild(rs_GlobalP rs_global, z_NodeP father,
			 int k, int l, uint32 *num)
{
  /* Find and allocate next child
     or return NULL if none exists.
     Search from k (incl) to l (excl).
     Return NULL if k == l.
     l may be -1.
     */
  z_NodeP next, f_next;
  int step;
  z_rs_ProcSetP app_ps;

  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);
  if (app_ps->rs_NextChild)
    return app_ps->rs_NextChild(app_global, father, k, l, num);

  step = (l > k) ? 1 : -1;
  for (; k != l; k += step) {
    next = app_ps->rs_AdjOracle(app_global, father, k);
    if (next) {
      if (!app_ps->rs_IsRoot(app_global, next)) {
	/* next is not the root of the reverse tree */
	f_next = app_ps->rs_F(app_global, next);
	if (app_ps->rs_Equal(app_global, father, f_next)) {
				/* found a child of father */
	  z_FreeNode(f_next);
	  *num = k;
	  return next;
	}
	z_FreeNode(f_next);
      }
      z_FreeNode(next);
    } /* (next) */
  } /* for */
  return NULL;			/* there are no more children */
} /* NextChild */

static uint32 FindChildNumber(rs_GlobalP rs_global,
			      z_NodeP son, z_NodeP father)
{
  /* Find the son's number */
  unsigned int k;
  z_NodeP other;
  z_rs_ProcSetP app_ps;

  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);

  if (app_ps->rs_FindChildNumber)
    return app_ps->rs_FindChildNumber(app_global, son, father);
  
  for (k=0;; k++) {		/* find number of son as a son of father */
    assert(k<app_ps->rs_Delta(app_global, father));
    other = app_ps->rs_AdjOracle(app_global, father, k);
    if (other) {
      if (app_ps->rs_Equal(app_global, other, son))
	/* found number of this as a son of father */
	break;
      z_FreeNode(other);
    }
  } /* for */
  z_FreeNode(other);
  return k;
} /* FindChildNumber */

static void rs_ForwardWithCache(rs_GlobalP rs_global,
				z_NodeP *nodeH, uint32 *k)
{
  z_NodeP child;
  z_rs_ProcSetP app_ps;
  
  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);
  child = *nodeH;
  /* child has a father */
  if (cacheSize) {
    cacheHits ++;
    assert((unsigned long)child->sibling % 2);
    *nodeH = cacheBottom;
    *k = (unsigned long)child->sibling / 2;
    cacheBottom = cacheBottom->parent;
    cacheSize --;
    if (cacheSize)
      cacheBottom->child = NULL;
    else
      cacheBottom = cacheTop = NULL;
    z_FreeNode(child);
  } else {
    cacheMisses ++;
    if (app_ps->rs_Forward)
      app_ps->rs_Forward(app_global, nodeH, k);
    else {
      *nodeH = app_ps->rs_F(app_global, child);
      *k = FindChildNumber(rs_global, child, *nodeH);
      z_FreeNode(child);
    }
    (*nodeH)->sibling = NULL;
  }
} /* rs_ForwardWithCache */

static void EmptyCache(void)
{
  z_NodeP h;

  while (cacheSize) {
    h = cacheTop;
    cacheTop = h->child;
    z_FreeNode(h);
    cacheSize --;
  }
  assert(! cacheTop);
  cacheBottom = NULL;
} /* EmptyCache */

void z_rs_Fini(void)
{
  if (0 && z_rs_cacheSize)
    printf("[%2d]Cache: size=%d, hits=%ld, misses=%ld, (%d)\n",
	   z_myproc, z_rs_cacheSize, cacheHits, cacheMisses, cacheSize);
  EmptyCache();
} /* z_rs_Fini */

void z_rs_Next(rs_GlobalP rs_global, z_rs_WorkUnitP *workUnitH)
{
  z_rs_ProcSetP app_ps;
  uint32 k, l;
  z_NodeP next, father, h;
  boolean isCommon;

  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);
  assert((*workUnitH)->depth_start >= (*workUnitH)->depth_common);
  father = (*workUnitH)->start;
  assert(father);
  if (!app_ps->rs_Condition || app_ps->rs_Condition(app_global, father))
    if (app_ps->rs_Result) app_ps->rs_Result(app_global, father);
  k = 0;

  if (cacheSize && !father->sibling)
    EmptyCache();		/* when load balancing transfers nodes */
				/* between processors, it clears sibling */
  
  while (1) {
    isCommon = (*workUnitH)->depth_start == (*workUnitH)->depth_common;
    l = isCommon ? (*workUnitH)->nr_end : app_ps->rs_Delta(app_global, father);
    assert(k <= l);
    next = NextChild(rs_global, father,
		     k, l, &k);
    if (next) {
      if (z_rs_cacheSize) {		/* put father into cache */
	next->sibling = (z_NodeP) (long) (1 + 2*k);
	father->parent = cacheBottom;
	father->child = NULL;
	if (cacheBottom) cacheBottom->child = father;
	cacheBottom = father;
	if (cacheSize < z_rs_cacheSize) {
	  if (cacheSize == 0) cacheTop = father;
	  cacheSize ++;
	} else {
	  h = cacheTop;
	  cacheTop = h->child;
	  cacheTop->parent = NULL;
	  z_FreeNode(h);
	}
      } else
	z_FreeNode(father);
      (*workUnitH)->start = next;
      (*workUnitH)->depth_start ++;
      break;
    }
    /* father has no more children */
    if (isCommon) {
      z_FreeNode(father);		/* father was the last node */
      (*workUnitH)->start = NULL;
      z_FreeNode((z_NodeP) *workUnitH);
      *workUnitH = NULL;
      return;
    }
    assert(! app_ps->rs_IsRoot(app_global, father));
    /* father has no more children, but it has a father */
    rs_ForwardWithCache(rs_global, &father, &k);
    (*workUnitH)->depth_start --;
    k++;
  } /* while (1) */
} /* z_rs_Next */

#ifndef SEQUENTIAL
static z_NodeP MiddleChild(rs_GlobalP rs_global,
			   z_NodeP father,
			   int k, int l,
			   uint32 *left_nr_end, uint32 *right_nr_end)
{
  /* Find and allocate brother in ]k,l[, if possible */	/* k < l */
  /* k >= -1, l >= 0 */
  z_NodeP split;
  int mid;

  if (l > 0) {
    mid = (k + l) / 2;		/* k <= mid < l */
    assert(mid < l);
    split = NextChild(rs_global, father, mid+1, l, left_nr_end);/* [mid+1,l[ */
    if (split) {
      *left_nr_end = mid+1;
      *right_nr_end = l;
      return split;
    }
    split = NextChild(rs_global, father, mid, k, left_nr_end); /* ]k,mid] */
    *right_nr_end = *left_nr_end+1;
    return split;
  } else
    return NULL;
} /* MiddleChild */

static z_NodeP rs_FindBrother(rs_GlobalP rs_global,
				z_NodeP this,
				uint32 *left_nr_end, uint32 *right_nr_end,
				uint32 nr_end,
				int32 depth_this,
				int32 *right_depth_start,
				int32 *depth_common)
{
  /* Pre: this is on the path: common < this <= start.
     Find and allocate a brother of this or a brother of an ancestor of this.
     */
  z_NodeP split, father;
  unsigned int k, l;
  z_rs_ProcSetP app_ps;

  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);

  assert(! app_ps->rs_IsRoot(app_global, this));

  father = app_ps->rs_F(app_global, this);

  if (depth_this - 1 == *depth_common) {
    /* father == common: Find brother between this and end or return NULL */
    l = nr_end;
  } else {
    /* 1. Split further up the tree (recursive call) or
       2. Find right brother of this or
       3. Return NULL
       */
    split = rs_FindBrother(rs_global, father,
			     left_nr_end, right_nr_end, nr_end,
			     depth_this - 1, right_depth_start,
			     depth_common);
    if (split) {
      z_FreeNode(father);
      return split;
    }
    l = app_ps->rs_Delta(app_global, father);
  }
  k = FindChildNumber(rs_global, this, father);
  /* Find brother in ]k,l[ */	/* k < l */
  split = MiddleChild(rs_global, father, k, l, left_nr_end, right_nr_end);
  if (split) {
    *right_depth_start = depth_this;
    *depth_common = depth_this - 1;
  }
  z_FreeNode(father);
  return split;
} /* rs_FindBrother */

void z_rs_Split(rs_GlobalP rs_global,
		z_rs_WorkUnitP workUnit,
		s_seqP workList)
{
  /* Cut the right part from workUnit and store both in workList */
  z_NodeP right_start;
  z_rs_WorkUnitP right_wu;
  z_rs_ProcSetP app_ps;
  unsigned int l;
  uint32 left_nr_end, right_nr_end;
  int32 right_depth_start=-555;

  assert(!s_seq_numElements(workList));
  app_ps = (z_rs_ProcSetP) z_GetProcSet(rs_global->rs_psn);
  assert(workUnit->start);
  s_seq_InsertLast(workList, workUnit);
  right_nr_end = workUnit->nr_end;

  if (! (workUnit->depth_start == workUnit->depth_common)) {
				/* end is not a child of start */
    if (cacheSize && !workUnit->start->sibling)
      EmptyCache();		/* when load balancing transfers nodes */
				/* between processors, it clears sibling */
  
    right_start = rs_FindBrother(rs_global,
				   workUnit->start,
				   &left_nr_end, &right_nr_end,
				   workUnit->nr_end,
				   workUnit->depth_start,
				   &right_depth_start,
				   &workUnit->depth_common);
    l = app_ps->rs_Delta(app_global, workUnit->start);
  } else {			/* end is a child of start */
    right_start = NULL;
    l = workUnit->nr_end;
  }
  if (!right_start) {
    /* right_start == NULL
       -> we need a child of start with number less than l */
				/* try children of start */
    right_start = MiddleChild(rs_global, workUnit->start, -1, l,
			      &left_nr_end, &right_nr_end);
    if (!right_start) {
      /* Interval has size 1 and cannot be split. */
      /* Update common, nr_end... */
      return;
    }
    workUnit->depth_common = workUnit->depth_start;
    right_depth_start = workUnit->depth_common + 1;
  }
  right_start->sibling = NULL;
  assert(! app_ps->rs_Equal(app_global, right_start, workUnit->start));
  /*printf("[%2d]splitted\n", z_myproc);
    app_ps->rs_Result(app_global, workUnit->start);
    app_ps->rs_Result(app_global, split);*/
  right_wu = NewWorkUnit(right_start, right_nr_end,
			 right_depth_start, workUnit->depth_common);
  s_seq_InsertFirst(workList, right_wu);
  workUnit->nr_end = left_nr_end;
} /* z_rs_Split */
#endif /* SEQUENTIAL */

#ifndef PEonly			/* front end only */

void z_rs_Restart(z_ProcSetNum procsetnum,
		  z_lb_GlobalP ap_global,
		  boolean sequential)
{
  rs_GlobalP rs_global;

  assert(rs_inited);
  rs_global = NewRsGlobal(procsetnum, ap_global);
  z_lb_RestartWork(lb_procsetnum, (z_lb_GlobalP)rs_global, sequential);
} /* z_rs_Restart */

#endif /* ! PEonly */

void z_rs_Enumerate(z_ProcSetNum procsetnum,
		    z_lb_GlobalP global,
		    z_NodeP root,
		    boolean sequential)
{
  /* Enumerate the subtree starting at root. Call rs_Result(node) for
     every node where rs_Condition(node) returns true. */
  z_rs_ProcSetP app_ps;
  rs_GlobalP rs_global;
  z_rs_WorkUnitP workUnit;

  assert(rs_inited);
  assert(!cacheTop && !cacheBottom);
  app_ps = (z_rs_ProcSetP) z_GetProcSet(procsetnum);
  assert(app_ps->rs_Equal);
  assert(app_ps->rs_F);
  assert(app_ps->rs_AdjOracle ||
	 (app_ps->rs_NextChild && app_ps->rs_FindChildNumber));
  assert(app_ps->rs_Delta);
#ifdef SEQUENTIAL
  sequential = 1;
#endif /* SEQUENTIAL */
  rs_global = NewRsGlobal(procsetnum, global);
  workUnit = NewWorkUnit(root, app_ps->rs_Delta(global, root), 0, 0);
  z_lb_Work(lb_procsetnum, (z_lb_GlobalP)rs_global, workUnit, sequential);
} /* z_rs_Enumerate */


/* Initialization ------------------------------------------------------ */

void z_rs_Init(void)
/* Initialize this module */
{
  if (!rs_inited) {
    rs_inited = 1;
    z_lb_Init();

    rs_GlobalTag = z_InstallClass(rs_Global, 1, 0, 0);
    z_rs_WorkUnitTag = z_InstallClass(z_rs_WorkUnit, 1, 0, 0);
    lb_procset.lb_Work = (z_lb_WorkProc) z_rs_Next;
    lb_procset.lb_WorkAndSplit = NULL;
#ifdef SEQUENTIAL
    lb_procset.lb_Split = NULL;
#else /* SEQUENTIAL */
    lb_procset.lb_Split = (z_lb_SplitProc) z_rs_Split;
#endif /* SEQUENTIAL */
    lb_procset.lb_Fini = z_rs_Fini;
    lb_procsetnum = z_InstallProcSet((z_ProcSetP) &lb_procset);
    cacheTop = cacheBottom = NULL;
  }
} /* z_rs_Init */

/* The ancestor cache only works if the load balancing algorithm does not
   switch between work units.

   Should broadcast z_rs_cacheSize at start.
   */
